#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2014 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
The algorithmic part of :term:`capturing`.

Prefer accessing the public members from the parent module.
"""

from __future__ import unicode_literals

import logging
from string import ascii_uppercase

from future.builtins import str

import numpy as np

from . import Coords, _parse
from ._parse import Cell


log = logging.getLogger(__name__)

try:
    from xlrd import colname as xl_colname
    # TODO: Try different backends providing `colname` function.
except ImportError:
    log.warning(
        'One of `xlrd`, `...` libraries is needed, will crash later!')


CHECK_CELLTYPE = False
"""When `True`, most coord-functions accept any 2-tuples."""


class EmptyCaptureException(Exception):
    """
    Thrown when :term:`targeting` fails.
    """

_special_coord_symbols = {'^', '_', '.'}

_primitive_dir_vectors = {
    'L': Coords(0, -1),
    'U': Coords(-1, 0),
    'R': Coords(0, 1),
    'D': Coords(1, 0)
}


def coords2Cell(row, col):
    """Make *A1* :class:`Cell` from *resolved* coords, with rudimentary error-checking.

    Examples::

        >>> coords2Cell(row=0, col=0)
        Cell(row='1', col='A')
        >>> coords2Cell(row=0, col=26)
        Cell(row='1', col='AA')

        >>> coords2Cell(row=10, col='.')
        Cell(row='11', col='.')

        >>> coords2Cell(row=-3, col=-2)
        Traceback (most recent call last):
        AssertionError: negative row!


    """
    if row not in _special_coord_symbols:
        assert row >= 0, 'negative row!'
        row = str(row + 1)
    if col not in _special_coord_symbols:
        assert col >= 0, 'negative col!'
        col = xl_colname(col)
    return Cell(row=row, col=col)


def _row2num(coord):
    """
    Resolves special coords or converts Excel 1-based rows to zero-based, reporting invalids.

    :param str, int coord:     excel-row coordinate or one of ``^_.``
    :return:    excel row number, >= 0
    :rtype:     int

    Examples::

        >>> row = _row2num('1')
        >>> row
        0
        >>> row == _row2num(1)
        True

    Negatives (from bottom) are preserved::

        >>> _row2num('-1')
        -1

    Fails ugly::

        >>> _row2num('.')
        Traceback (most recent call last):
        ValueError: invalid literal for int() with base 10: '.'
    """
    rcoord = int(coord)
    if rcoord == 0:
        msg = 'Uncooked-coord cannot be zero!'
        raise ValueError(msg.format(coord))
    if rcoord > 0:
        rcoord -= 1

    return rcoord


def _col2num(coord):
    """
    Resolves special coords or converts Excel A1 columns to a zero-based, reporting invalids.

    :param str coord:          excel-column coordinate or one of ``^_.``
    :return:    excel column number, >= 0
    :rtype:     int

    Examples::

        >>> col = _col2num('D')
        >>> col
        3
        >>> _col2num('d') == col
        True
        >>> _col2num('AaZ')
        727
        >>> _col2num('10')
        9
        >>> _col2num(9)
        8

    Negatives (from left-end) are preserved::

        >>> _col2num('AaZ')
        727

    Fails ugly::

        >>> _col2num('%$')
        Traceback (most recent call last):
        ValueError: substring not found

        >>> _col2num([])
        Traceback (most recent call last):
        TypeError: int() argument must be a string, a bytes-like object or
                    a number, not 'list'

    """
    try:
        rcoord = int(coord)
    except ValueError:
        rcoord = 0
        for c in coord:
            rcoord = rcoord * 26 + ascii_uppercase.rindex(c.upper()) + 1

        rcoord -= 1
    else:
        if rcoord == 0:
            msg = 'Uncooked-coord cannot be zero!'
            raise ValueError(msg.format(coord))
        elif rcoord > 0:
            rcoord -= 1

    return rcoord


def _resolve_coord(cname, cfunc, coord, up_coord, dn_coord, base_coords=None):
    """
    Translates special coords or converts Excel string 1-based rows/cols to zero-based, reporting invalids.

    :param str cname:
            the coord-name, one of 'row', 'column'
    :param function cfunc:
            the function to convert coord ``str --> int``
    :param int, str coord:
            the "A1" coord to translate
    :param int up_coord:
            the resolved *top* or *left* margin zero-based coordinate
    :param int dn_coord:
            the resolved *bottom* or *right* margin zero-based coordinate
    :param int, None base_coords:
            the resolved basis for dependent coord, if any

    :return: the resolved coord or `None` if it were not a special coord.


    Row examples::

        >>> cname = 'row'

        >>> r0 = _resolve_coord(cname, _row2num, '1', 1, 10)
        >>> r0
        0
        >>> r0 == _resolve_coord(cname, _row2num, 1, 1, 10)
        True
        >>> _resolve_coord(cname, _row2num, '^', 1, 10)
        1
        >>> _resolve_coord(cname, _row2num, '_', 1, 10)
        10
        >>> _resolve_coord(cname, _row2num, '.', 1, 10, 13)
        13
        >>> _resolve_coord(cname, _row2num, '-3', 0, 10)
        8

    But notice when base-cell missing::

        >>> _resolve_coord(cname, _row2num, '.', 0, 10, base_coords=None)
        Traceback (most recent call last):
        ValueError: Cannot resolve `relative-row` without `base-coord`!

    Other ROW error-checks::

        >>> _resolve_coord(cname, _row2num, '0', 0, 10)
        Traceback (most recent call last):
        ValueError: invalid row('0') due to: Uncooked-coord cannot be zero!

        >>> _resolve_coord(cname, _row2num, 'a', 0, 10)
        Traceback (most recent call last):
        ValueError: invalid row('a') due to: invalid literal for int() with base 10: 'a'

        >>> _resolve_coord(cname, _row2num, None, 0, 10)
        Traceback (most recent call last):
        ValueError: invalid row(None) due to:
                int() argument must be a string,
                a bytes-like object or a number, not 'NoneType'


    Column examples::

        >>> cname = 'column'

        >>> _resolve_coord(cname, _col2num, 'A', 1, 10)
        0
        >>> _resolve_coord(cname, _col2num, 'DADA', 1, 10)
        71084
        >>> _resolve_coord(cname, _col2num, '.', 1, 10, 13)
        13
        >>> _resolve_coord(cname, _col2num, '-4', 0, 10)
        7

    And COLUMN error-checks::

        >>> _resolve_coord(cname, _col2num, None, 0, 10)
        Traceback (most recent call last):
        ValueError: invalid column(None) due to: int() argument must be a string,
                    a bytes-like object or a number, not 'NoneType'

        >>> _resolve_coord(cname, _col2num, 0, 0, 10)
        Traceback (most recent call last):
        ValueError: invalid column(0) due to: Uncooked-coord cannot be zero!

    """
    try:
        if coord in _special_coord_symbols:
            special_dict = {
                '^': up_coord,
                '_': dn_coord
            }
            if base_coords is not None:
                special_dict['.'] = base_coords
            rcoord = special_dict[coord]
        else:
            rcoord = cfunc(coord)

            # Resolve negatives as from the end.
            if rcoord < 0:
                rcoord = dn_coord + rcoord + 1

        return rcoord
    except Exception as ex:
        if isinstance(ex, KeyError) and ex.args == ('.',):
            msg = "Cannot resolve `relative-{}` without `base-coord`!"
            raise ValueError(msg.format(cname))
        msg = 'invalid {}({!r}) due to: {}'
        # fututils.raise_from(ValueError(msg.format(cname, coord, ex)), ex) see
        # GH 141
        raise ValueError(msg.format(cname, coord, ex))


def _resolve_cell(cell, up_coords, dn_coords, base_coords=None):
    """
    Translates any special coords to absolute ones.

    To get the margin_coords, use one of:

    * :meth:`ABCSheet.get_margin_coords()`
    * :func:`.io._sheets.margin_coords_from_states_matrix()`

    :param Cell cell:
            The "A1" cell to translate its coords.
    :param Coords up_coords:
            the top-left resolved coords with full-cells
    :param Coords dn_coords:
            the bottom-right resolved coords with full-cells
    :param Coords base_coords:
                A resolved cell to base dependent coords (``.``).
    :return: the resolved cell-coords
    :rtype:  Coords


    Examples::

        >>> up = Coords(1, 2)
        >>> dn = Coords(10, 6)
        >>> base = Coords(40, 50)

        >>> _resolve_cell(Cell(col='B', row='5'), up, dn)
        Coords(row=4, col=1)

        >>> _resolve_cell(Cell('^', '^'), up, dn)
        Coords(row=1, col=2)

        >>> _resolve_cell(Cell('_', '_'), up, dn)
        Coords(row=10, col=6)

        >>> base == _resolve_cell(Cell('.', '.'), up, dn, base)
        True

        >>> _resolve_cell(Cell('-1', '-2'), up, dn)
        Coords(row=10, col=5)

        >>> _resolve_cell(Cell('A', 'B'), up, dn)
        Traceback (most recent call last):
        ValueError: invalid cell(Cell(row='A', col='B')) due to:
                invalid row('A') due to: invalid literal for int() with base 10: 'A'

    But notice when base-cell missing::

        >>> _resolve_cell(Cell('1', '.'), up, dn)
        Traceback (most recent call last):
        ValueError: invalid cell(Cell(row='1', col='.')) due to:
        Cannot resolve `relative-col` without `base-coord`!

    """
    assert not CHECK_CELLTYPE or isinstance(cell, Cell), cell
    assert not CHECK_CELLTYPE or isinstance(up_coords, Coords), up_coords
    assert not CHECK_CELLTYPE or isinstance(dn_coords, Coords), dn_coords
    try:
        if base_coords is None:
            base_row = base_col = None
        else:
            base_row, base_col = base_coords
        row = _resolve_coord('row', _row2num, cell.row,
                             up_coords[0], dn_coords[0], base_row)
        col = _resolve_coord('col', _col2num, cell.col,
                             up_coords[1], dn_coords[1], base_col)

        return Coords(row, col)
    except Exception as ex:
        msg = "invalid cell(%r) due to: %s\n  margins(%r)\n  base_coords(%r)"
        log.debug(msg, cell, ex, (up_coords, dn_coords), base_coords)
        msg = "invalid cell(%r) due to: %s"
        # fututils.raise_from(ValueError(msg % (cell, ex)), ex) see GH 141
        raise ValueError(msg % (cell, ex))


_mov_vector_slices = {
    # VECTO_SLICE        REVERSE  COORD_INDEX
    'L': (1, -1, lambda r, c: (r, slice(None, c + 1))),
    'U': (0, -1, lambda r, c: (slice(None, r + 1), c)),
    'R': (1, 1, lambda r, c: (r, slice(c, None))),
    'D': (0, 1, lambda r, c: (slice(r, None), c)),
}


def _extract_states_vector(states_matrix, dn_coords, land, mov):
    """Extract a slice from the states-matrix by starting from `land` and following `mov`."""
    coord_indx, is_reverse, slice_func = _mov_vector_slices[mov]
    vect_slice = slice_func(*land)
    states_vect = states_matrix[vect_slice]
    states_vect = states_vect[::is_reverse]

    return states_vect, coord_indx, is_reverse


def _target_opposite(states_matrix, dn_coords, land, moves, edge_name=''):
    """
    Follow moves from `land` and stop on the 1st full-cell.

    :param np.ndarray states_matrix:
            A 2D-array with `False` wherever cell are blank or empty.
            Use :meth:`ABCSheet.get_states_matrix()` to derrive it.
    :param Coords dn_coords:
            the bottom-right for the top-left of full-cells
    :param Coords land:
            the landing-cell
    :param str moves: MUST not be empty
    :return: the identified target-cell's coordinates
    :rtype: Coords


    Examples::

        >>> states_matrix = np.array([
        ...     [0, 0, 0, 0, 0, 0],
        ...     [0, 0, 0, 0, 0, 0],
        ...     [0, 0, 0, 1, 1, 1],
        ...     [0, 0, 1, 0, 0, 1],
        ...     [0, 0, 1, 1, 1, 1]
        ... ])
        >>> args = (states_matrix, Coords(4, 5))

        >>> _target_opposite(*(args + (Coords(0, 0), 'DR')))
        Coords(row=3, col=2)

        >>> _target_opposite(*(args + (Coords(0, 0), 'RD')))
        Coords(row=2, col=3)

    It fails if a non-empty target-cell cannot be found, or
    it ends-up beyond bounds::

        >>> _target_opposite(*(args + (Coords(0, 0), 'D')))
        Traceback (most recent call last):
        pandalone.xleash._capture.EmptyCaptureException: No opposite-target found
                        while moving(D) from landing-Coords(row=0, col=0)!

        >>> _target_opposite(*(args + (Coords(0, 0), 'UR')))
        Traceback (most recent call last):
        pandalone.xleash._capture.EmptyCaptureException: No opposite-target found
                        while moving(UR) from landing-Coords(row=0, col=0)!


    But notice that the landing-cell maybe outside of bounds::

        >>> _target_opposite(*(args + (Coords(3, 10), 'L')))
        Coords(row=3, col=5)

    """
    assert not CHECK_CELLTYPE or isinstance(dn_coords, Coords), dn_coords
    assert not CHECK_CELLTYPE or isinstance(land, Coords), land

    up_coords = np.array([0, 0])
    target = np.array(land)

    if land[0] > dn_coords[0] and 'U' in moves:
        target[0] = dn_coords[0]
    if land[1] > dn_coords[1] and 'L' in moves:
        target[1] = dn_coords[1]

#     if states_matrix[target].all():
#         return Coords(*target)

    imoves = iter(moves)
    mov1 = next(imoves)
    mov2 = next(imoves, None)
    dv2 = mov2 and _primitive_dir_vectors[mov2]

    # Limit negative coords, since they are valid indices.
    while (up_coords <= target).all():
        try:
            states_vect, coord_indx, is_reverse = _extract_states_vector(
                states_matrix, dn_coords, target, mov1)
        except IndexError:
            break
        else:
            if states_vect.any():
                indices = states_vect.nonzero()[0]
                target[coord_indx] += is_reverse * indices.min()

                return Coords(*target)

            if not dv2:
                break

            target += dv2

    msg = 'No opposite-target found while moving({}) from {}landing-{}!'
    raise EmptyCaptureException(msg.format(moves, edge_name, land))


def _target_same_vector(states_matrix, dn_coords, land, mov):
    """
    :param np.ndarray states_matrix:
            A 2D-array with `False` wherever cell are blank or empty.
            Use :meth:`ABCSheet.get_states_matrix()` to derrive it.
    :param Coords dn_coords:
            the bottom-right for the top-left of full-cells
    :param Coords land:
            The landing-cell, which MUST be full!
    """
    states_vect, coord_indx, is_reverse = _extract_states_vector(
        states_matrix, dn_coords, land, mov)
    if states_vect.all():
        same_len = len(states_vect) - 1
    else:
        indices = np.diff(states_vect).nonzero()[0]
        same_len = indices.min()
    target_coord = land[coord_indx] + is_reverse * same_len

    return target_coord, coord_indx


def _target_same(states_matrix, dn_coords, land, moves, edge_name=''):
    """
    Scan term:`exterior` row and column on specified `moves` and stop on the last full-cell.

    :param np.ndarray states_matrix:
            A 2D-array with `False` wherever cell are blank or empty.
            Use :meth:`ABCSheet.get_states_matrix()` to derrive it.
    :param Coords dn_coords:
            the bottom-right for the top-left of full-cells
    :param Coords land:
            the landing-cell which MUST be within bounds
    :param moves: which MUST not be empty
    :return: the identified target-cell's coordinates
    :rtype: Coords


    Examples::

        >>> states_matrix = np.array([
        ...     [0, 0, 0, 0, 0, 0],
        ...     [0, 0, 0, 0, 0, 0],
        ...     [0, 0, 0, 1, 1, 1],
        ...     [0, 0, 1, 0, 0, 1],
        ...     [0, 0, 1, 1, 1, 1]
        ... ])
        >>> args = (states_matrix, Coords(4, 5))

        >>> _target_same(*(args + (Coords(4, 5), 'U')))
        Coords(row=2, col=5)

        >>> _target_same(*(args + (Coords(4, 5), 'L')))
        Coords(row=4, col=2)

        >>> _target_same(*(args + (Coords(4, 5), 'UL', )))
        Coords(row=2, col=2)


    It fails if landing is empty or beyond bounds::

        >>> _target_same(*(args + (Coords(2, 2), 'DR')))
        Traceback (most recent call last):
        pandalone.xleash._capture.EmptyCaptureException: No same-target found
                        while moving(DR) from landing-Coords(row=2, col=2)!

        >>> _target_same(*(args + (Coords(10, 3), 'U')))
        Traceback (most recent call last):
        pandalone.xleash._capture.EmptyCaptureException: No same-target found
                        while moving(U) from landing-Coords(row=10, col=3)!

    """
    assert not CHECK_CELLTYPE or isinstance(dn_coords, Coords), dn_coords
    assert not CHECK_CELLTYPE or isinstance(land, Coords), land

    target = np.asarray(land)
    if (target <= dn_coords).all() and states_matrix[land]:
        for mov in moves:
            coord, indx = _target_same_vector(states_matrix, dn_coords,
                                              np.asarray(land), mov)
            target[indx] = coord

        return Coords(*target)
    msg = 'No same-target found while moving({}) from {}landing-{}!'
    raise EmptyCaptureException(msg.format(moves, edge_name, land))


def _sort_rect(r1, r2):
    """
    Sorts rect-vertices in a 2D-array (with vertices in rows).

    Example::

        >>> _sort_rect((5, 3), (4, 6))
        array([[4, 3],
               [5, 6]])
    """
    rect = np.array([r1, r2], dtype=int)
    rect.sort(0)
    return rect


def _expand_rect(states_matrix, r1, r2, exp_moves):
    """
    Applies the :term:`expansion-moves` based on the `states_matrix`.

    :param state:
    :param Coords r1:
              any vertice of the rect to expand
    :param Coords r2:
              any vertice of the rect to expand
    :param np.ndarray states_matrix:
            A 2D-array with `False` wherever cell are blank or empty.
            Use :meth:`ABCSheet.get_states_matrix()` to derrive it.
    :param exp_moves:
            Just the parsed string, and not `None`.
    :return: a sorted rect top-left/bottom-right


    Examples::

        >>> states_matrix = np.array([
        ...     #0  1  2  3  4  5
        ...     [0, 0, 0, 0, 0, 0], #0
        ...     [0, 0, 1, 1, 1, 0], #1
        ...     [0, 1, 0, 0, 1, 0], #2
        ...     [0, 1, 1, 1, 1, 0], #3
        ...     [0, 0, 0, 0, 0, 1], #4
        ... ], dtype=bool)

        >>> r1, r2 = (Coords(2, 1), Coords(2, 1))
        >>> _expand_rect(states_matrix, r1, r2, 'U')
        (Coords(row=2, col=1), Coords(row=2, col=1))

        >>> r1, r2 = (Coords(3, 1), Coords(2, 1))
        >>> _expand_rect(states_matrix, r1, r2, 'R')
        (Coords(row=2, col=1), Coords(row=3, col=4))

        >>> r1, r2 = (Coords(2, 1), Coords(6, 1))
        >>> _expand_rect(states_matrix, r1, r2, 'r')
        (Coords(row=2, col=1), Coords(row=6, col=5))

        >>> r1, r2 = (Coords(2, 3), Coords(2, 3))
        >>> _expand_rect(states_matrix, r1, r2, 'LURD')
        (Coords(row=1, col=1), Coords(row=3, col=4))

    """
    assert not CHECK_CELLTYPE or isinstance(r1, Coords), r1
    assert not CHECK_CELLTYPE or isinstance(r2, Coords), r2

    exp_moves = _parse.parse_expansion_moves(exp_moves)

    nd_offsets = np.array([0, 1, 0, 1])
    coord_offsets = {
        'L': np.array([0,  0, -1, 0]),
        'R': np.array([0,  0,  0, 1]),
        'U': np.array([-1, 0,  0, 0]),
        'D': np.array([0,  1,  0, 0]),
    }
    coord_indices = {
        'L': [0, 1, 2, 2],
        'R': [0, 1, 3, 3],
        'U': [0, 0, 2, 3],
        'D': [1, 1, 2, 3],
    }

    # Sort rect's vertices top-left/bottom-right.
    #
    rect = _sort_rect(r1, r2)
    # ``[r1, r2, c1, c2]`` to use slices, below
    rect = rect.T.flatten()
    for dirs_repeated in exp_moves:
        for dirs in dirs_repeated:
            orig_rect = rect
            for d in dirs:
                exp_rect = rect + coord_offsets[d]
                exp_vect_i = exp_rect[coord_indices[d]] + nd_offsets
                exp_vect_v = states_matrix[slice(*exp_vect_i[:2]),
                                           slice(*exp_vect_i[2:])]
                if exp_vect_v.any():
                    rect = exp_rect
            if (rect == orig_rect).all():
                break

    return Coords(*rect[[0, 2]]), Coords(*rect[[1, 3]])


def resolve_capture_rect(states_matrix, up_dn_margins,
                         st_edge, nd_edge=None, exp_moves=None,
                         base_coords=None):
    """
    Performs :term:`targeting`, :term:`capturing` and :term:`expansions` based on the :term:`states-matrix`.

    To get the margin_coords, use one of:

    * :meth:`ABCSheet.get_margin_coords()`
    * :func:`.io._sheets.margin_coords_from_states_matrix()`

    Its results can be fed into :func:`read_capture_values()`.

    :param np.ndarray states_matrix:
            A 2D-array with `False` wherever cell are blank or empty.
            Use :meth:`ABCSheet.get_states_matrix()` to derrive it.
    :param (Coords, Coords) up_dn_margins:
            the top-left/bottom-right coords with full-cells
    :param Edge st_edge: "uncooked" as matched by regex
    :param Edge nd_edge: "uncooked" as matched by regex
    :param list or none exp_moves:
            Just the parsed string, and not `None`.
    :param Coords base_coords:
            The base for a :term:`dependent` :term:`1st` edge.

    :return:    a ``(Coords, Coords)`` with the 1st and 2nd :term:`capture-cell`
                ordered from top-left --> bottom-right.
    :rtype: tuple

    :raises EmptyCaptureException:
            When :term:`targeting` failed, and no :term:`target` cell identified.

    Examples::
        >>> from pandalone.xleash import Edge, margin_coords_from_states_matrix

        >>> states_matrix = np.array([
        ...     [0, 0, 0, 0, 0, 0],
        ...     [0, 0, 0, 0, 0, 0],
        ...     [0, 0, 0, 1, 1, 1],
        ...     [0, 0, 1, 0, 0, 1],
        ...     [0, 0, 1, 1, 1, 1]
        ... ], dtype=bool)
        >>> up, dn = margin_coords_from_states_matrix(states_matrix)

        >>> st_edge = Edge(Cell('1', 'A'), 'DR')
        >>> nd_edge = Edge(Cell('.', '.'), 'DR')
        >>> resolve_capture_rect(states_matrix, (up, dn), st_edge, nd_edge)
        (Coords(row=3, col=2), Coords(row=4, col=2))

    Using dependenent coordinates for the 2nd edge::

        >>> st_edge = Edge(Cell('_', '_'), None)
        >>> nd_edge = Edge(Cell('.', '.'), 'UL')
        >>> rect = resolve_capture_rect(states_matrix, (up, dn), st_edge, nd_edge)
        >>> rect
        (Coords(row=2, col=2), Coords(row=4, col=5))

    Using sheet's margins::

        >>> st_edge = Edge(Cell('^', '_'), None)
        >>> nd_edge = Edge(Cell('_', '^'), None)
        >>> rect == resolve_capture_rect(states_matrix, (up, dn), st_edge, nd_edge)
        True

    Walking backwards::

        >>> st_edge = Edge(Cell('^', '_'), 'L')          # Landing is full, so 'L' ignored.
        >>> nd_edge = Edge(Cell('_', '_'), 'L', '+')    # '+' or would also stop.
        >>> rect == resolve_capture_rect(states_matrix, (up, dn), st_edge, nd_edge)
        True

    """
    up_margin, dn_margin = up_dn_margins
    assert not CHECK_CELLTYPE or isinstance(up_margin, Coords), up_margin
    assert not CHECK_CELLTYPE or isinstance(dn_margin, Coords), dn_margin

    st = _resolve_cell(st_edge.land, up_margin, dn_margin, base_coords)
    try:
        st_state = states_matrix[st]
    except IndexError:
        st_state = False

    if st_edge.mov is not None:
        if st_state:
            if st_edge.mod == '+':
                st = _target_same(states_matrix, dn_margin, st, st_edge.mov,
                                  '1st-')
        else:
            st = _target_opposite(states_matrix, dn_margin, st, st_edge.mov,
                                  '1st-')

    if nd_edge is None:
        nd = None
    else:
        nd = _resolve_cell(nd_edge.land, up_margin, dn_margin, st)

        if nd_edge.mov is not None:
            try:
                nd_state = states_matrix[nd]
            except IndexError:
                nd_state = False

            mov = nd_edge.mov
            if nd_state:
                if (nd_edge.mod == '+' or
                        nd_edge.land == Cell('.', '.') and nd_edge.mod != '?'):
                    nd = _target_same(
                        states_matrix, dn_margin, nd, mov, '2nd-')
            else:
                nd = _target_opposite(
                    states_matrix, dn_margin, nd, mov, '2nd-')

    if exp_moves:
        st, nd = _expand_rect(states_matrix, st, nd or st, exp_moves)
    else:
        if nd is not None:
            rect = _sort_rect(st, nd)
            st, nd = tuple(Coords(*c) for c in rect)

    return st, nd



#! python
#-*- coding: utf-8 -*-
#
# Copyright 2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

from contextlib import contextmanager
from io import StringIO
import logging
import os
import re
import sys
from tempfile import mkdtemp
from textwrap import dedent
import unittest


log = logging.getLogger(__name__)


DEFAULT_LOG_LEVEL = logging.INFO


# TODO: Make it public
def init_logging(module_name, loglevel=DEFAULT_LOG_LEVEL):
    logging.basicConfig(level=loglevel)
    logging.getLogger().setLevel(level=loglevel)

    log = logging.getLogger(module_name)
    log.trace = lambda *args, **kws: log.log(0, *args, **kws)

    return log

##############
#  Compatibility
#
try:  # pragma: no cover
    assertRaisesRegex = unittest.TestCase.assertRaisesRegex
except:  # pragma: no cover
    assertRaisesRegex = unittest.TestCase.assertRaisesRegexp


class CustomAssertions(object):
    # Idea from: http://stackoverflow.com/a/15868615/548792

    def _raise_file_msg(self, msg, path, user_msg):
        msg = msg % os.path.abspath(path)
        if user_msg:
            msg = '%s: %s' % (msg, user_msg)
        raise AssertionError(msg)

    def assertFileExists(self, path, user_msg=None):
        if not os.path.lexists(path):
            self._raise_file_msg(
                "File does not exist in path '%s'!", path, user_msg)

    def assertFileNotExists(self, path, user_msg=None):
        if os.path.lexists(path):
            self._raise_file_msg("File exists in path '%s'!", path, user_msg)

    def assertNumpyEqual(self, x, y, msg=''):
        """
        Adapted from :func:`numpy.testing.utils.assert_array_compare()`.
        """
        import numpy as np
        from numpy import testing as npt

        def _isXXX(a, op, when_ex):
            res = []
            for i in a.flatten():
                try:
                    t = op(i)
                except TypeError:
                    t = when_ex
                res.append(t)
            return np.array(res, dtype=bool)

        def _isNaN(a):
            return _isXXX(a, np.isnan, True)

        def _isInf(a):
            return _isXXX(a, np.isinf, False)

        def _isPInf(a):
            return _isXXX(a, lambda x: x == np.Inf, False)

        def _isMInf(a):
            return _isXXX(a, lambda x: x == -np.Inf, False)

        cond = (x.shape == () and y.shape == ()) or x.shape == y.shape
        if not cond:
            msg = npt.build_err_msg([x, y],
                                    msg
                                    + '\n(shapes %s, %s mismatch)' % (x.shape,
                                                                      y.shape))
        if x.dtype != y.dtype:
            msg = npt.build_err_msg([x, y],
                                    msg
                                    + '\n(dtypes %s, %s mismatch)' % (x.dtype,
                                                                      y.dtype))
            raise AssertionError(msg)
        x_isnan, y_isnan = _isNaN(x), _isNaN(y)
        x_isinf, y_isinf = _isInf(x), _isInf(y)

        # Validate that the special values are in the same place
        if any(x_isnan) or any(y_isnan):
            npt.assert_array_equal(x_isnan, y_isnan, 'nan: ' + msg)
        if any(x_isinf) or any(y_isinf):
            # Check +inf and -inf separately, since they are different
            npt.assert_array_equal(_isPInf(x), _isPInf(y), '+inf: ' + msg)
            npt.assert_array_equal(_isMInf(x), _isMInf(y), '-inf: ' + msg)

        # Combine all the special values
        x_id, y_id = x_isnan, y_isnan
        x_id |= x_isinf
        y_id |= y_isinf

        # Only do the comparison if actual values are left
        if all(x_id):
            return

        if any(x_id):
            val = np.equal(x[~x_id], y[~y_id])
        else:
            val = np.equal(x, y)

        if isinstance(val, bool):
            cond = val
            reduced = [0]
        else:
            reduced = val.ravel()
            cond = reduced.all()
            reduced = reduced.tolist()

        if not cond:
            match = 100 - 100.0 * reduced.count(1) / len(reduced)
            msg = npt.build_err_msg([x, y],
                                    msg
                                    + '\n(mismatch %s%%)' % (match,))
            if not cond:
                raise AssertionError(msg)

    def assertStrippedStringsEqual(self, st, nd, msg=None, context_chars=30):
        regex = re.compile('\\s+', re.DOTALL)
        st1 = regex.sub('', st)
        nd1 = regex.sub('', nd)
        if st1 != nd1:
            err_i = len(os.path.commonprefix((st1, nd1)))
            s_slice = slice(max(0, err_i - context_chars),
                            err_i + context_chars)
            c1, c2 = st1[s_slice], nd1[s_slice]
            frmt = dedent("""\
            Stripped-strings differ at char %i (lens: 1st=%i, 2nd=%s)!
              --1st: %s
                     %s^
              --2nd: %s
              ==1st original: %s
              ==2nd original: %s
            ----%s
            """)
            spcs = ' ' * context_chars
            err_msg = frmt % (err_i, len(st1), len(nd1), c1, spcs, c2,
                              st, nd, (msg or ''))
            self.fail(err_msg)

    def assertStrippedStringsStartsWith(self, st, nd, msg=None, context_chars=30):
        regex = re.compile('\\s+', re.DOTALL)
        st1 = regex.sub('', st)
        nd1 = regex.sub('', nd)
        if not st1.startswith(nd1):
            err_i = len(os.path.commonprefix((st1, nd1)))
            s_slice = slice(max(0, err_i - context_chars),
                            err_i + context_chars)
            c1, c2 = st1[s_slice], nd1[s_slice]
            frmt = dedent("""\
            Stripped-strings differ at char %i (lens: 1st=%i, 2nd=%s)!
              --1st: %s
                     %s^
              --2nd: %s
              ==1st original: %s
              ==2nd original: %s
            ----%s
            """)
            spcs = ' ' * context_chars
            err_msg = frmt % (err_i, len(st1), len(nd1), c1, spcs, c2,
                              st, nd, (msg or ''))
            self.fail(err_msg)

try:
    from tempfile import TemporaryDirectory  # @UnusedImport
except ImportError:
    class TemporaryDirectory(object):

        """Create and return a temporary directory.  This has the same
        behavior as mkdtemp but can be used as a context manager.  For
        example:

            with TemporaryDirectory() as tmpdir:
                ...

        Upon exiting the context, the directory and everything contained
        in it are removed.

        From: http://stackoverflow.com/questions/19296146/tempfile-temporarydirectory-context-manager-in-python-2-7
              http://stackoverflow.com/a/19299884/548792
        """

        def __init__(self, suffix="", prefix="tmp",
                     dir=None):  # @ReservedAssignment
            self._closed = False
            self.name = None  # Handle mkdtemp raising an exception
            self.name = mkdtemp(suffix, prefix, dir)

        def __repr__(self):
            return "<{} {!r}>".format(self.__class__.__name__, self.name)

        def __enter__(self):
            return self.name

        def cleanup(self, _warn=False):
            if self.name and not self._closed:
                try:
                    self._rmtree(self.name)
                except (TypeError, AttributeError) as ex:
                    # Issue #10188: Emit a warning on stderr
                    # if the directory could not be cleaned
                    # up due to missing globals
                    if "None" not in str(ex):
                        raise
                    log.warning("ERROR: %s while cleaning up %s", ex, self)
                    return
                self._closed = True
                log.warning("Implicitly cleaning up %s", self)

        def __exit__(self, exc, value, tb):
            self.cleanup()

        def __del__(self):
            # Issue a ResourceWarning if implicit cleanup needed
            self.cleanup(_warn=True)

        # XXX (ncoghlan): The following code attempts to make
        # this class tolerant of the module nulling out process
        # that happens during CPython interpreter shutdown
        # Alas, it doesn't actually manage it. See issue #10188
        _listdir = staticmethod(os.listdir)
        _path_join = staticmethod(os.path.join)
        _isdir = staticmethod(os.path.isdir)
        _islink = staticmethod(os.path.islink)
        _remove = staticmethod(os.remove)
        _rmdir = staticmethod(os.rmdir)

        def _rmtree(self, path):
            # Essentially a stripped down version of shutil.rmtree.  We can't
            # use globals because they may be None'ed out at shutdown.
            for name in self._listdir(path):
                fullname = self._path_join(path, name)
                try:
                    isdir = self._isdir(
                        fullname) and not self._islink(fullname)
                except OSError:
                    isdir = False
                if isdir:
                    self._rmtree(fullname)
                else:
                    try:
                        self._remove(fullname)
                    except OSError:
                        pass
            try:
                self._rmdir(path)
            except OSError:
                pass


@contextmanager
def capture(command, *args, **kwargs):
    # Unused
    out, sys.stdout = sys.stdout, StringIO()
    err, sys.stderr = sys.stderr, StringIO()
    try:
        command(*args, **kwargs)
        sys.stdout.seek(0)
        yield (sys.stdout.getvalue(), sys.stderr.getvalue())
    finally:
        sys.stdout = out
        sys.stderr = err


@contextmanager
def chdir(dirname=None):
    curdir = os.getcwd()
    try:
        os.chdir(dirname)
        yield
    finally:
        os.chdir(curdir)


def xw_close_workbook(wb):
    try:
        app = wb.app
        wb.close()
        if not app.books:
            # TODO: Workaround
            # https://github.com/ZoomerAnalytics/xlwings/issues/548
            app.quit()
    except Exception:
        log.warning('Minor failure while closing Workbook!', exc_info=True)


@contextmanager
def xw_no_save_Workbook(wb_name=None):
    import xlwings as xw

    wb = xw.Book(wb_name)
    try:
        yield wb
    finally:
        xw_close_workbook(wb)
    # app.quit()



#!/usr/bin/env python

""" Git Versioning Script

Will transform stdin to expand some keywords with git version/author/date information.

Specify --clean to remove this information before commit.

Setup:

1. [DONE] Setup sources::

    git add bin/gitstamping.py
    echo '*.py filter=gitstamping' >> .gitattributes
 

2. [DONE] Add a within files this statements::

    __commit__ = ""


3. Setup your working-dir::

    git config filter.gitstamping.smudge 'python bin/gitstamping.py'
    git config filter.gitstamping.clean  'python bin/gitstamping.py --clean'


From https://gist.github.com/pkrusche/7369262
"""

import re
import subprocess
import sys


def filter_or_smudge(clean):
    rexp1 = re.compile(r'__commit__(\s*)=(\s*)".*')
    if clean:
        for line in sys.stdin:
            line = re.sub(rexp1, r'__commit__\1=\2""', line)
            sys.stdout.write(re.sub(rexp1, r'__commit__\1=\2""', line))
    else:
        git_id = subprocess.check_output(['git', 'describe', '--always'])
        git_id = git_id.decode(encoding='utf_8')
        git_id = re.sub(r'[\n\r\t"\']', "", git_id)
        for line in sys.stdin:
            line = re.sub(rexp1, r'__commit__\1=\2"%s"' % git_id, line)
            sys.stdout.write(line)


def main(args):
    clean = len(sys.argv) > 1 and sys.argv[1] == '--clean'
    filter_or_smudge(clean)

if __name__ == "__main__":
    main(sys.argv)



#! python
#-*- coding: utf-8 -*-
#
# Copyright 2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

from __future__ import division, unicode_literals

import doctest
import sys
import unittest

from pandalone.components import (Assembly, FuncComponent, Pstep)
import pandalone.components
from pandalone.mappings import pmods_from_tuples
from tests._tutils import init_logging


log = init_logging(__name__)


@unittest.skipIf(sys.version_info < (3, 4), "Doctests are made for py >= 3.3")
class Doctest(unittest.TestCase):

    def test_doctests(self):
        failure_count, test_count = doctest.testmod(
            pandalone.components,
            optionflags=doctest.NORMALIZE_WHITESPACE)  # | doctest.ELLIPSIS)
        self.assertGreater(test_count, 0, (failure_count, test_count))
        self.assertEquals(failure_count, 0, (failure_count, test_count))


class TestComponents(unittest.TestCase):

    def test_assembly(self):
        def cfunc_f1(comp, value_tree):
            comp.pinp().A
            comp.pout().B

        def cfunc_f2(comp, value_tree):
            comp.pinp().B
            comp.pout().C
        ass = Assembly(FuncComponent(cfunc) for cfunc in [cfunc_f1, cfunc_f2])
        ass._build()
        self.assertEqual(list(ass._iter_validations()), [])
        self.assertEqual(ass._inp, ['f1/A', 'f2/B'])
        self.assertEqual(ass._out, ['f1/B', 'f2/C'])

        pmods = pmods_from_tuples([('f1', r'/root1'), ('f2', r'/root2')])
        ass._build(pmods)
        self.assertEqual(
            sorted(ass._inp + ass._out),
            ['/root1/A', '/root1/B', '/root2/B', '/root2/C'])


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



#! python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
''''Installation script for *pandalone*.


Install:
========
To install it, assuming you have download the sources,
do the usual::

    python setup.py install

Or get it directly from the PIP repository::

    pip install pandalone

Or get it directly from the github repository::

    pip install git+https://github.com/pandalone/pandalone.git
'''
# Got ideas for project-setup from many places, among others:
#    http://www.jeffknupp.com/blog/2013/08/16/open-sourcing-a-python-project-the-right-way/
#    http://python-packaging-user-guide.readthedocs.org/en/latest/current.html

import io
import os
import re
import sys

from setuptools import setup, find_packages


__commit__ = ""

# Fail early on ancient python-versions
#
py_ver = sys.version_info
if py_ver < (2, 7):
    exit("Sorry, Python2 >= 2.7 is supported!")
if py_ver >= (3,) and py_ver < (3, 3):
    exit("Sorry, Python3 >= 3.3 is supported!")

proj_name = 'pandalone'
mydir = os.path.dirname(__file__)


# Version-trick to have version-info in a single place,
# taken from: http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
##
def read_project_version():
    fglobals = {}
    with io.open(os.path.join(
            mydir, proj_name, '_version.py'), encoding='UTF-8') as fd:
        exec(fd.read(), fglobals)  # To read __version__
    return fglobals['__version__']


def read_text_lines(fname):
    with io.open(os.path.join(mydir, fname)) as fd:
        return fd.readlines()


def yield_rst_only_markup(lines):
    """
    :param file_inp:     a `filename` or ``sys.stdin``?
    :param file_out:     a `filename` or ``sys.stdout`?`

    """
    # TODO: Change role substitution from Sphinx black-list to RsT white-list.
    substs = [
        # Selected Sphinx-only Roles.
        #
        (r':abbr:`([^`]+)`',        r'\1'),
        (r':ref:`([^`]+)`',         r'ref: *\1*'),
        (r':term:`([^`]+)`',        r'**\1**'),
        (r':dfn:`([^`]+)`',         r'**\1**'),
        (r':(samp|guilabel|menuselection|doc|file|func|mod|class'
         r'|program|envvar|command):`([^`]+)`', r'\1``\2``'),


        # Sphinx-only roles:
        #        :foo:`bar`   --> foo(``bar``)
        #        :a:foo:`bar` XXX afoo(``bar``)
        #
        #(r'(:(\w+))?:(\w+):`([^`]*)`', r'\2\3(``\4``)'),
        #(r':(\w+):`([^`]*)`', r'\1(`\2`)'),


        # Sphinx-only Directives.
        #
        (r'\.\. doctest',           r'code-block'),
        (r'\.\. plot::',            r'.. '),
        (r'\.\. seealso',           r'info'),
        (r'\.\. glossary',          r'rubric'),
        (r'\.\. figure::',          r'.. '),
        (r'\.\. image::',          r'.. '),


        # Other
        #
        (r'\|version\|',              r'x.x.x'),
    ]

    regex_subs = [(re.compile(regex, re.IGNORECASE), sub)
                  for (regex, sub) in substs]

    def clean_line(line):
        try:
            for (regex, sub) in regex_subs:
                line = regex.sub(sub, line)
        except Exception as ex:
            print("ERROR: %s, (line(%s)" % (regex, sub))
            raise ex

        return line

    for line in lines:
        yield clean_line(line)


proj_ver = read_project_version()


readme_lines = read_text_lines('README.rst')
description = readme_lines[1]
long_desc = ''.join(yield_rst_only_markup(readme_lines))
# Trick from: http://peterdowns.com/posts/first-time-with-pypi.html
download_url = 'https://github.com/%s/%s/tarball/v%s' % (
    proj_name, proj_name, proj_ver)

install_requires = [
    'future >= 0.15.0',
    'toolz',
    'jsonschema >= 2.5',
    'numpy',
    'openpyxl>=2.2',
    'easygui != 0.98',  # For excel-macros locked msg-box.
]
if not os.environ.get('READTHEDOCS') == 'True':
    install_requires.append('sphinx_rtd_theme')
install_requires.append('asteval>=0.9.7'  # https://github.com/newville/asteval/issues/16
                        if py_ver[:2] >= (3, 5) else 'asteval')

setup(
    name=proj_name,
    version=proj_ver,
    description=description,
    long_description=long_desc,
    author="Kostis Anagnostopoulos at European Commission (JRC)",
    author_email="ankostis@gmail.com",
    url="https://github.com/%s/%s" % (proj_name, proj_name),
    download_url=download_url,
    keywords=[
        "python", "utility", "library", "data", "tree", "processing",
        "calculation", "dependencies", "resolution", "scientific",
        "engineering", "pandas", "simulink",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: Implementation :: CPython",
        "Development Status :: 4 - Beta",
        'Natural Language :: English',
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Manufacturing",
        'Environment :: Console',
        "License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    packages=find_packages(exclude=["tests"]),
    #include_package_data = True,
    package_data={
        proj_name: ['excel/*.vba', 'excel/*.ico', 'icons/*'],
    },
    install_requires=install_requires,
    setup_requires=[
        'setuptools',
    ],
    # build_requires=[
    #     # PEP426-field actually not used by `pip` them, hence
    #     # included in /requirements/developmnet.pip.
    #     'setuptools',
    #     'setuptools-git >= 0.3',  # Gather package-data from all files in git.
    #     'wheel',
    #     'collective.checkdocs',
    # ],
    # dev_requires=[
    #     # PEP426-field actually not used by `pip` them, hence
    #     # included in /requirements/developmnet.pip.
    #     'sphinx>=1.4',
    # ],
    tests_require=[
        'ddt',
        'nose',
        'coverage',
        'coveralls',
        'pandas',
        'xlrd',
    ],
    test_suite='nose.collector',
    extras_require={
        ':python_version == "2.7"': [  # See PEP-426
            'mock',
        ],
        'xlwings': ['xlwings >= 0.9.2'],
        'pandas': ['pandas >= 0.18'],  # For xleash pandas filter.
        'xlrd': ['xlrd'],
    },
    entry_points={
        'console_scripts': [
            'pndl = %s.__main__:main' % proj_name,
        ],
        'pandalone.xleash.plugins': [
            'xlrd_be = pandalone.xleash.io._xlrd:load_as_xleash_plugin [xlrd]',
            'pandas_filters = pandalone.xleash._pandas_filters:load_as_xleash_plugin [pandas]',
        ],
    },
    zip_safe=True,
    options={
        'build_sphinx': {
            'build_dir': 'doc/_build',
        },
        'bdist_wheel': {
            'universal': True,
        },
    },
    platforms=['any'],
    # TODO: check when `run_requires` works.
    # run_requires=[
    #     {
    #         "requires": ["pywin32 > 1.0", "xlwings", "easygui"],
    #         "environment": "sys_platform == 'win32' or sys_platform == 'darwin'"
    #     },
    # ],
    # metadata_version="2.0",
)



#! python
# -*- coding: UTF-8 -*-
#
# Copyright 2015 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl

import doctest
import os
import re
import subprocess
import sys
import unittest

import pandalone

import os.path as osp
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch  # @UnusedImport


mydir = osp.dirname(__file__)
proj_path = osp.normpath(osp.join(mydir, '..'))
readme_path = osp.join(proj_path, 'README.rst')
tutorial_path = osp.join(proj_path, 'doc', 'tutorial.rst')


@unittest.skipIf(sys.version_info < (3, 4), "Doctests are made for py >= 3.3")
class Doctest(unittest.TestCase):

    def test_doctest_README(self):
        failure_count, test_count = doctest.testfile(
            readme_path, module_relative=False,
            optionflags=doctest.NORMALIZE_WHITESPACE)
        self.assertGreater(test_count, 0, (failure_count, test_count))
        self.assertEquals(failure_count, 0, (failure_count, test_count))

    def test_README_version_opening(self):
        ver = pandalone.__version__
        header_len = 20
        mydir = osp.dirname(__file__)
        with open(readme_path) as fd:
            for i, l in enumerate(fd):
                if ver in l:
                    break
                elif i >= header_len:
                    msg = "Version(%s) not found in README %s header-lines!"
                    raise AssertionError(msg % (ver, header_len))

    def test_README_reldate_opening(self):
        reldate = pandalone.__updated__
        header_len = 20
        mydir = osp.dirname(__file__)
        with open(readme_path) as fd:
            for _, l in zip(range(header_len), fd):
                if reldate in l:
                    break
            else:
                msg = "Version(%s) not found in README first %s header-lines!"
                raise AssertionError(msg % (reldate, header_len))

    def test_README_version_cmdline(self):
        ver = pandalone.__version__
        mydir = osp.dirname(__file__)
        with open(readme_path) as fd:
            ftext = fd.read()
            m = re.search(
                r'pndlcmd --version\s+%s' % ver, ftext, re.MULTILINE | re.IGNORECASE)
            self.assertIsNotNone(m,
                                 "Version(%s) not found in README cmd-line version-check!" %
                                 ver)

    def test_README_as_PyPi_landing_page(self):
        from docutils import core as dcore

        long_desc = subprocess.check_output(
            'python setup.py --long-description'.split(),
            cwd=proj_path)
        self.assertIsNotNone(long_desc, 'Long_desc is null!')

        with patch('sys.exit'):
            dcore.publish_string(long_desc, enable_exit_status=False,
                                 settings_overrides={  # see `docutils.frontend` for more.
                                     'halt_level': 2  # 2=WARN, 1=INFO
                                 })
